package com.muyu.datasources.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.muyu.common.core.utils.StringUtils;
import com.muyu.common.domain.DataSources;
import com.muyu.common.domain.Structure;
import com.muyu.common.domain.TableInfo;
import com.muyu.common.domain.req.DataSourcesReq;
import com.muyu.common.pool.MysqlPool;
import com.muyu.common.pool.OraclePool;
import com.muyu.common.pool.RedisPool;
import com.muyu.common.security.utils.SecurityUtils;
import com.muyu.datasources.mapper.DataSourcesMapper;
import com.muyu.datasources.service.DataSourcesService;
import com.muyu.datasources.service.StructureService;
import com.muyu.datasources.service.TableInfoService;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import redis.clients.jedis.Jedis;

import java.sql.*;
import java.util.Date;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import static com.muyu.common.pool.base.BaseConfig.*;

/**
 * @Author：zhangzhihao
 * @name：DataSourcesServiceImpl
 * @Date：2024/8/21 14:35
 * 不准抄代码，添加注释，清楚每一行代码意思
 */
@Log4j2
@Service
public class DataSourcesServiceImpl extends ServiceImpl<DataSourcesMapper, DataSources> implements DataSourcesService {


    @Autowired
    private DataSourcesMapper dataSourcesMapper;

    @Autowired
    private TableInfoService tableInfoService;

    @Autowired
    private StructureService structureService;


    private static final Long PARENTID =0L;
    private static final String DRIVERNAME ="com.mysql.cj.jdbc.Driver";
    private static final String DRIVERORACLENAME ="oracle.jdbc.driver.OracleDriver";
    private static final String MYSQL ="mysql";
    private static final String ORACLE ="oracle";
    private static final String REDIS ="redis";

    /**
     * 添加数据来源
     * @param dataSources
     * @return
     */
    @Override
    public boolean save(DataSources dataSources) {

        String type = dataSources.getType();

        if (type.equals(MYSQL)){
            dataSources.setDriverName(DRIVERNAME);
        }
        if (type.equals(ORACLE)){
            dataSources.setDriverName(DRIVERORACLENAME);
        }

        return super.save(dataSources);
    }





    /**
     * 分页查询所有数据
     * @param req
     * @return
     */
    @Override
    public Page<DataSources> selectList(DataSourcesReq req) {

        LambdaQueryWrapper<DataSources> queryWrapper = new LambdaQueryWrapper<>();
        queryWrapper.like(StringUtils.isNotEmpty(req.getName()), DataSources::getName,req.getName());
        queryWrapper.like(StringUtils.isNotEmpty(req.getDatabaseName()), DataSources::getDatabaseName,req.getDatabaseName());

        Page<DataSources> page = new Page<>(req.getPageNum(), req.getPageSize());


        Page<DataSources> pageList = dataSourcesMapper.selectPage(page, queryWrapper);


        return pageList;
    }

    /**
     * 测试连接
     * @param dataSources
     */
    @Override
    public int connectDataSources(DataSources dataSources) {
        if (dataSources.getType().equals(MYSQL)){
            MysqlPool mysqlPool = new MysqlPool(dataSources);
            mysqlPool.init();
            Connection conn = mysqlPool.getConn();
            mysqlPool.replease(conn);
            mysqlPool.closeConn();
            return 1;
        }else if (dataSources.getType().equals(REDIS)){
            RedisPool redisPool = new RedisPool(dataSources);
            redisPool.init();
            Jedis conn = redisPool.getConn();
            redisPool.replease(conn);
            redisPool.closeConn();
            return 1;
        }
        return 0;
    }

    private static final ExecutorService threadPool = Executors.newCachedThreadPool();
    @Override
    public Integer syncAssetStructure(DataSources dataSources) {

        try {
            if (dataSources.getType().equals(MYSQL)) {
                MysqlPool mysqlPool = new MysqlPool(dataSources);
                mysqlPool.init();
                Connection conn = mysqlPool.getConn();


                TableInfo tableInfoInsert = TableInfo.builder()
                        .basicId(dataSources.getId())
                        .parentId(PARENTID)
                        .tableRemark("")
                        .isCenter("Y")
                        .type("dataSource")
                        .tableName(dataSources.getDatabaseName())
                        .createBy(SecurityUtils.getUsername())
                        .createTime(new Date())
                        .build();

                //添加数据库table_info表
                tableInfoService.saveOrUpdate(tableInfoInsert,new LambdaUpdateWrapper<TableInfo>(TableInfo.class){{
                    eq(TableInfo::getTableName,tableInfoInsert.getTableName());
                    eq(TableInfo::getBasicId,dataSources.getId());
                }});

                //根据数据库id和数据库名称或表名称查询
                TableInfo tableInfo = tableInfoService.selectTableInfoByName(tableInfoInsert);

                // 查询所有表信息
                DatabaseMetaData metaData = conn.getMetaData();
                ResultSet rs = metaData.getTables(dataSources.getDatabaseName(), null, "%", new String[]{"TABLE"});

                while (rs.next()) {
                    String tableName = rs.getString("TABLE_NAME");
                    String tableRemark = rs.getString("REMARKS");

                    try (PreparedStatement preparedStatement = conn.prepareStatement(SELECTCOUNT + tableName)) {
                        ResultSet resultSet = preparedStatement.executeQuery();
                        long count = 0L;
                        if (resultSet.next()) {
                            count = resultSet.getLong("count");
                        }

                        TableInfo build = TableInfo.builder()
                                .basicId(dataSources.getId())
                                .tableName(tableName)
                                .tableRemark(tableRemark != null ? tableRemark : "")
                                .parentId(tableInfo.getId())
                                .type("dataTable")
                                .isCenter("Y")
                                .updateBy(SecurityUtils.getUsername())
                                .dataNum(count)
                                .updateTime(new Date())
                                .build();
                        tableInfoService.saveOrUpdate(build, new LambdaUpdateWrapper<>(TableInfo.class) {{
                            eq(TableInfo::getTableName, build.getTableName());
                            eq(TableInfo::getBasicId, dataSources.getId());
                        }});
                        TableInfo table = tableInfoService.selectTableInfoByName(build);

                        threadPool.submit(() -> {
                            syncData(conn, dataSources.getDatabaseName(), table);
                        });
                    } catch (SQLException e) {
                        log.error("查询计数失败", e);
                    }
                }

                try {
                    Thread.sleep(1000);
                    conn.close();
                    mysqlPool.closeConn();

                } catch (SQLException e) {
                    log.error("关闭数据库连接失败", e);
                } catch (InterruptedException e) {
                    throw new RuntimeException(e);
                }
            } else if (dataSources.getType().equals(ORACLE)) {
                OraclePool oraclePool = new OraclePool(dataSources);
                oraclePool.init();
                Connection conn = oraclePool.getConn();

                TableInfo tableInfoInsert = TableInfo.builder()
                        .basicId(dataSources.getId())
                        .parentId(PARENTID)
                        .tableRemark("")
                        .isCenter("Y")
                        .type("dataSource")
                        .tableName(dataSources.getDatabaseName())
                        .createBy(SecurityUtils.getUsername())
                        .createTime(new Date())
                        .build();

                // 添加数据库 table_info 表
                tableInfoService.saveOrUpdate(tableInfoInsert, new LambdaUpdateWrapper<>(TableInfo.class) {{
                    eq(TableInfo::getTableName, tableInfoInsert.getTableName());
                    eq(TableInfo::getBasicId, dataSources.getId());
                }});

                // 根据数据库 id 和数据库名称查询
                TableInfo tableInfo = tableInfoService.selectTableInfoByName(tableInfoInsert);

                DatabaseMetaData metaData = conn.getMetaData();

                // 编写 SQL 查询语句，从 ZZH 模式的 T_USER 表中查询 NAME 字段
                String sql = "SELECT TABLE_NAME FROM DBA_TABLES WHERE OWNER = '" + dataSources.getDatabaseName() + "'";

                // 执行查询
                PreparedStatement prepared = conn.prepareStatement(sql);
                ResultSet rs = prepared.executeQuery();

                // 遍历 ResultSet 对象，获取 NAME 字段的值
                while (rs.next()) {
                    String tableName = rs.getString(1);

                    try (PreparedStatement preparedStatement = conn.prepareStatement(SELECTCOUNT + tableName)) {
                        ResultSet resultSet = preparedStatement.executeQuery();
                        long count = 0L;
                        if (resultSet.next()) {
                            count = resultSet.getLong("count");
                        }

                        TableInfo build = TableInfo.builder()
                                .basicId(dataSources.getId())
                                .tableName(tableName)
                                .tableRemark("")
                                .parentId(tableInfo.getId())
                                .type("dataTable")
                                .isCenter("Y")
                                .updateBy(SecurityUtils.getUsername())
                                .dataNum(count)
                                .updateTime(new Date())
                                .build();
                        tableInfoService.saveOrUpdate(build, new LambdaUpdateWrapper<>(TableInfo.class) {{
                            eq(TableInfo::getTableName, build.getTableName());
                            eq(TableInfo::getBasicId, dataSources.getId());
                        }});
                        TableInfo table = tableInfoService.selectTableInfoByName(build);

                        threadPool.submit(() -> {
                            syncOracleData(conn, dataSources.getDatabaseName(), table);
                        });
                    } catch (SQLException e) {
                        log.error("查询计数失败", e);
                    }
                }

                try {

                    Thread.sleep(1000);
                    conn.close();
                    oraclePool.closeConn();
                } catch (SQLException e) {
                    log.error("关闭数据库连接失败", e);
                } catch (InterruptedException e) {
                    throw new RuntimeException(e);
                }
            }
        } catch (SQLException e) {
            log.error("数据库操作失败", e);
            throw new RuntimeException(e);
        }
        return 1;
    }


    /**
     * 同步数据库表结构信息
     *
     * @param conn 数据库连接
     * @param databaseName 数据库名称
     * @param table 表信息，包含表名和表ID
     */
    public void syncData(Connection conn, String databaseName, TableInfo table) {
        // 创建一个单线程执行器，用于提交保存或更新结构信息的任务
        ExecutorService threadPool = Executors.newSingleThreadExecutor();

        // 准备SQL查询语句，用于获取指定数据库表的列信息
        String sql = "SELECT\n" +
                "    COLUMN_NAME,\n" +
                "    COLUMN_COMMENT,\n" +
                "    CASE WHEN COLUMN_KEY = 'PRI' THEN '是' ELSE '否' END AS IS_PRIMARY_KEY,\n" +
                "    CASE\n" +
                "        WHEN DATA_TYPE = 'int' THEN 'Integer'\n" +
                "        WHEN DATA_TYPE = 'bigint' THEN 'Long'\n" +
                "        WHEN DATA_TYPE = 'varchar' THEN 'String'\n" +
                "        WHEN DATA_TYPE = 'decimal' THEN 'BigDecimal'\n" +
                "        WHEN DATA_TYPE = 'tinyint' AND COLUMN_TYPE = 'tinyint(1)' THEN 'Boolean'\n" +
                "        ELSE DATA_TYPE\n" +
                "    END AS GENERIC_DATA_TYPE,\n" +
                "    DATA_TYPE AS ORIGINAL_DATA_TYPE,\n" +
                "    COLUMN_TYPE AS DETAILED_DATA_TYPE,\n" +
                "    CHARACTER_MAXIMUM_LENGTH,\n" +
                "    NUMERIC_SCALE,\n" +
                "    IS_NULLABLE,\n" +
                "    COLUMN_DEFAULT\n" +
                "FROM\n" +
                "    INFORMATION_SCHEMA.COLUMNS\n" +
                "WHERE\n" +
                "    TABLE_SCHEMA = ?\n" +
                "    AND TABLE_NAME = ?;";

        // 创建PreparedStatement对象，用于执行SQL查询
        PreparedStatement ps = null;
        // 创建ResultSet对象，用于存储查询结果
        ResultSet resultSet = null;

        try {
            // 准备SQL语句并设置参数
            ps = conn.prepareStatement(sql);
            ps.setString(1, databaseName);
            ps.setString(2, table.getTableName());

            // 执行查询并处理结果
            resultSet = ps.executeQuery();
            while (resultSet.next()) {
                // 获取列信息并构建Structure对象
                Structure build = Structure.builder()
                        .tableId(table.getId())
                        .columnName(resultSet.getString(1))
                        .columnRemark(resultSet.getString(2))
                        .isPrimary("是".equals(resultSet.getString(3)) ? "Y" : "N")
                        .javaType(resultSet.getString(4))
                        .columnType(resultSet.getString(6))
                        .columnLength(String.valueOf(resultSet.getInt(7)))
                        .columnDecimals(String.valueOf(resultSet.getInt(8)))
                        .isNull("YES".equals(resultSet.getString(9)) ? "Y" : "N")
                        .defaultValue(resultSet.getString(10))
                        .build();

                // 提交保存或更新结构信息的任务到线程池
                threadPool.submit(() -> {
                    try {
                        structureService.saveOrUpdate(build, new LambdaUpdateWrapper<Structure>() {{
                            eq(Structure::getTableId, build.getTableId());
                            eq(Structure::getColumnName, build.getColumnName());
                            eq(Structure::getColumnRemark, build.getColumnRemark());
                        }});
                    } catch (Exception e) {
                        log.error("保存或更新结构信息失败", e);
                    }
                });
            }

            // 关闭线程池并等待所有任务完成
            threadPool.shutdown();
            threadPool.awaitTermination(1, TimeUnit.MINUTES);
        } catch (SQLException e) {
            throw new RuntimeException("Error executing SQL query: " + e.getMessage(), e);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new RuntimeException("Thread interrupted", e);
        } finally {
            // 安静地关闭资源
            closeQuietly(ps);
            closeQuietly(resultSet);
            if (!threadPool.isTerminated()) {
                threadPool.shutdownNow();  // 强制关闭线程池
            }
        }
    }

    /**
     * 安静地关闭资源，忽略关闭时产生的异常
     *
     * @param closeable 需要关闭的资源
     */
    private void closeQuietly(AutoCloseable closeable) {
        if (closeable != null) {
            try {
                closeable.close();
            } catch (Exception e) {
                log.error("关闭资源时发生异常", e);
            }
        }
    }


    /**
     * 同步Oracle数据库表结构数据到本地
     *
     * @param conn 数据库连接
     * @param databaseName 数据库名称
     * @param table 表信息
     */
    public void syncOracleData(Connection conn, String databaseName, TableInfo table) {
        // 创建一个可缓存线程池，用于提交结构更新任务
        ExecutorService threadPool = Executors.newCachedThreadPool();
        // 准备执行SQL的PreparedStatement对象
        PreparedStatement ps = null;
        try {
            // 创建SQL，用于查询表的列信息，包括列名、数据类型等
            String sql = "SELECT  \n" +
                    "    COLUMN_NAME,  \n" +
                    "    NULL AS COLUMN_COMMENT,  \n" +
                    "    CASE  \n" +
                    "        WHEN DATA_TYPE = 'NUMBER' AND DATA_PRECISION IS NOT NULL AND DATA_SCALE = 0 THEN 'Integer'  \n" +
                    "        WHEN DATA_TYPE = 'NUMBER' THEN 'BigDecimal'  \n" +
                    "        WHEN DATA_TYPE = 'VARCHAR2' THEN 'String'  \n" +
                    "        WHEN DATA_TYPE = 'CHAR' THEN 'String'\n" +
                    "        WHEN DATA_TYPE = 'BINARY_FLOAT' OR DATA_TYPE = 'BINARY_DOUBLE' THEN 'Double' -- \n" +
                    "        WHEN DATA_TYPE = 'NUMBER' AND DATA_SCALE = 1 AND COLUMN_NAME LIKE '%IS_ACTIVE%' THEN 'Boolean' -- \n" +
                    "        ELSE DATA_TYPE  \n" +
                    "    END AS GENERIC_DATA_TYPE,  \n" +
                    "    DATA_TYPE AS ORIGINAL_DATA_TYPE,  \n" +
                    "    DATA_TYPE || '(' || CASE WHEN DATA_LENGTH IS NULL THEN NULL ELSE TO_CHAR(DATA_LENGTH) END  \n" +
                    "       || CASE WHEN DATA_SCALE IS NULL THEN NULL ELSE ', ' || TO_CHAR(DATA_SCALE) END || ')' AS DETAILED_DATA_TYPE,  \n" +
                    "    CASE WHEN DATA_TYPE IN ('VARCHAR2', 'CHAR') THEN DATA_LENGTH ELSE NULL END AS CHARACTER_MAXIMUM_LENGTH,    \n" +
                    "    NULLABLE AS IS_NULLABLE,  \n" +
                    "    DATA_DEFAULT AS COLUMN_DEFAULT  \n" +
                    "FROM  \n" +
                    "    ALL_TAB_COLUMNS  \n" +
                    "WHERE  \n" +
                    "    OWNER = UPPER(?) \n" +
                    "    AND TABLE_NAME = UPPER(?)";
            // 使用数据库连接预编译SQL
            ps = conn.prepareStatement(sql);
            // 设置SQL参数，数据库名称和表名
            ps.setString(1, databaseName);
            ps.setString(2, table.getTableName());

            // 执行查询并获取结果集
            ResultSet resultSet = ps.executeQuery();
            // 遍历结果集，处理每一行数据
            while (resultSet.next()) {
                // 从结果集中提取列信息
                String columnName = resultSet.getString(1);
                String columnComment = resultSet.getString(2);
                String javaType = resultSet.getString(3);
                String columnType = resultSet.getString(4);
                String isNullable = resultSet.getString(7);
                String columnDefault = resultSet.getString(8);

                // 构建表结构信息对象
                Structure build = Structure.builder()
                        .tableId(table.getId())
                        .columnName(columnName)
                        .columnRemark(columnComment)
                        .isPrimary("N")
                        .javaType(javaType)
                        .columnType(columnType)
                        .isNull("YES".equals(isNullable) ? "Y" : "N")
                        .defaultValue(columnDefault)
                        .build();

                // 提交任务到线程池，异步更新本地表结构信息
                threadPool.submit(() -> {
                    // 复制结构信息对象，准备更新
                    Structure buildCopy = Structure.builder()
                            .tableId(build.getTableId())
                            .columnName(build.getColumnName())
                            .columnRemark(build.getColumnRemark())
                            .isPrimary(build.getIsPrimary())
                            .javaType(build.getJavaType())
                            .columnType(build.getColumnType())
                            .isNull(build.getIsNull())
                            .defaultValue(build.getDefaultValue())
                            .build();

                    // 使用服务保存或更新结构信息
                    structureService.saveOrUpdate(buildCopy, new LambdaUpdateWrapper<Structure>() {{
                        eq(Structure::getTableId, buildCopy.getTableId());
                        eq(Structure::getColumnName, buildCopy.getColumnName());
                        eq(Structure::getColumnRemark, buildCopy.getColumnRemark());
                    }});
                });
            }

        } catch (SQLException e) {
            // 记录并抛出SQL异常
            log.error("同步Oracle数据时发生错误", e);
            throw new RuntimeException("同步Oracle数据时发生错误", e);
        } finally {
            // 关闭资源，先关闭PreparedStatement，然后关闭线程池
            if (ps != null) {
                try {
                    ps.close();
                } catch (SQLException e) {
                    log.error("关闭PreparedStatement出错", e);
                }
            }
            threadPool.shutdown();
        }
    }


//            else if (dataSources.getType().equals(REDIS)){
//                RedisPool redisPool = new RedisPool(dataSources);
//                redisPool.init();
//                Jedis jedis = redisPool.getConn();
//                String cursor = ScanParams.SCAN_POINTER_START;
//                ScanParams scanParams = new ScanParams().count(100);
//                HashMap<String, String> map = new HashMap<>();
//                while (true){
//                    ScanResult<String> scanResult = jedis.scan(cursor, scanParams);
//                    List<String> keys = scanResult.getResult();
//                    for (String key : keys) {
//                        String value = jedis.get(key);
//                        if (value!=null){
//                            map.put(key,value);
//                        }
//                    }
//                     cursor = scanResult.getCursor();
//                    if (cursor.equals(ScanParams.SCAN_POINTER_START)){
//                        break;
//                    }
//                }
//                System.out.println(map);
//                redisPool.replease(jedis);
//                redisPool.closeConn();
//            }


}
